--
-- ExtendingAxle
--
-- Authors: Xentro and Wopster
-- Description: Enables you to extend/withdraw vehicle axles based on a speed limit
--
-- Copyright (c) Wopster, 2017

ExtendingAxle = {}

ExtendingAxle.SPEED_THRESHOLD = 1
ExtendingAxle.EXTEND = 1
ExtendingAxle.WITHDRAW = -1

function ExtendingAxle.prerequisitesPresent(specializations)
    if not SpecializationUtil.hasSpecialization(AnimatedVehicle, specializations) then
        print("Warning: Specialization ExtendingAxle needs the specialization AnimatedVehicle")

        return false
    end

    return true
end

function ExtendingAxle:preLoad(savegame)
    self.extendAxle = ExtendingAxle.extendAxle
end

function ExtendingAxle:load(savegame)
    self.extendAxles = {
        minSpeed = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.extendingAxle#minSpeed"), 3),
        maxSpeedReachScale = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.extendingAxle#maxSpeedReachScale"), 10),
        maxSpeed = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.extendingAxle#maxSpeed"), 20),
        isActive = false,
        curSpeed = 0,
        curDir = ExtendingAxle.EXTEND,
        inputBinding = InputBinding.IMPLEMENT_EXTRA2
    }

    local extendAnimation = getXMLString(self.xmlFile, "vehicle.extendingAxle#animationName")
    if self.animations[extendAnimation] ~= nil then
        self.extendAxles.animation = extendAnimation
    end

    local input = getXMLString(self.xmlFile, "vehicle.extendingAxle#inputBinding")
    if input ~= nil and InputBinding[input] ~= nil then
        self.extendAxles.inputBinding = InputBinding[input]
    end
end

function ExtendingAxle:postLoad(savegame)
    if savegame ~= nil and not savegame.resetVehicles then
        local extendedAxleDir = getXMLInt(savegame.xmlFile, savegame.key .. "#extendedAxleDir")

        if extendedAxleDir ~= nil then
            self.extendAxles.curDir = extendedAxleDir

            if self.extendAxles.animation ~= nil then
                self:setAnimationTime(self.extendAxles.animation, extendedAxleDir, true)
            end
        end
    end
end

function ExtendingAxle:delete()
end

function ExtendingAxle:readStream(streamId, connection)
    self:extendAxle(streamReadInt8(streamId), true)
end

function ExtendingAxle:writeStream(streamId, connection)
    streamWriteInt8(streamId, self.extendAxles.curDir)
end

function ExtendingAxle:getSaveAttributesAndNodes(nodeIdent)
    local attributes

    attributes = ('extendedAxleDir="%s"'):format(self.extendAxles.curDir)

    return attributes, nil
end

function ExtendingAxle:mouseEvent(...)
end

function ExtendingAxle:keyEvent(...)
end

function ExtendingAxle:update(dt)
    if self:getIsActive() and self:getIsActiveForInput() and not self:hasInputConflictWithSelection() then
        if self.extendAxles.animation ~= nil then
            if InputBinding.hasEvent(self.extendAxles.inputBinding) then
                local dir = self.extendAxles.curDir == ExtendingAxle.EXTEND and ExtendingAxle.WITHDRAW or ExtendingAxle.EXTEND

                self:extendAxle(dir)
            end
        end
    end
end

function ExtendingAxle:updateTick(dt)
    if self:getIsActive() then
        if self.extendAxles.isActive and self.extendAxles.animation ~= nil then
            local lastSpeed = self:getLastSpeed()
            local speed = math.max(lastSpeed, 0)

            self.extendAxles.curSpeed = speed

            if speed >= self.extendAxles.minSpeed and speed <= self.extendAxles.maxSpeed then
                speed = math.min(speed / self.extendAxles.maxSpeedReachScale, 1)
            else
                speed = 0
            end

            self:setAnimationSpeed(self.extendAxles.animation, speed * self.extendAxles.curDir)
            self.extendAxles.isActive = self:getIsAnimationPlaying(self.extendAxles.animation)
        end
    end
end

function ExtendingAxle:draw()
    if g_i18n:hasText("action_extendAxle") and g_i18n:hasText("action_withdrawAxle") then
        local text = self.extendAxles.curDir < 0 and g_i18n:getText("action_withdrawAxle") or g_i18n:getText("action_extendAxle")

        if g_i18n:hasText("action_axleDirection") then
            g_currentMission:addHelpButtonText(string.format(g_i18n:getText("action_axleDirection"), text), self.extendAxles.inputBinding)
        end
    end

    if g_i18n:hasText("info_axleSpeeding") and g_i18n:hasText("info_axleToSlow") then
        if self.extendAxles.isActive then
            local lastSpeed = self:getLastSpeed()
            if lastSpeed >= self.extendAxles.maxSpeed then
                g_currentMission:showBlinkingWarning(g_i18n:getText("info_axleSpeeding"), 0)
            elseif lastSpeed <= self.extendAxles.minSpeed then
                g_currentMission:showBlinkingWarning(g_i18n:getText("info_axleToSlow"), 0)
            end
        end
    end
end

---
-- @param dir
-- @param noEventSend
--
function ExtendingAxle:extendAxle(dir, noEventSend)
    self.extendAxles.curDir = dir
    self.extendAxles.isActive = true

    if self.extendAxles.animation ~= nil then
        self:playAnimation(self.extendAxles.animation, dir, nil, true)
    end

    ExtendAxleEvent.sendEvent(self, dir, noEventSend)
end


ExtendAxleEvent = {}
ExtendAxleEvent_mt = Class(ExtendAxleEvent, Event)

InitEventClass(ExtendAxleEvent, "ExtendAxleEvent")

function ExtendAxleEvent:emptyNew()
    local event = Event:new(ExtendAxleEvent_mt)

    return event
end

function ExtendAxleEvent:new(object, dir)
    local event = ExtendAxleEvent:emptyNew()

    event.object = object
    event.dir = dir

    return event
end

function ExtendAxleEvent:readStream(streamId, connection)
    self.object = readNetworkNodeObject(streamId)
    self.dir = streamReadInt8(streamId)

    self:run(connection)
end

function ExtendAxleEvent:writeStream(streamId, connection)
    writeNetworkNodeObject(streamId, self.object)
    streamWriteInt8(streamId, self.dir)
end

function ExtendAxleEvent:run(connection)
    self.object:extendAxle(self.dir, true)

    if not connection:getIsServer() then
        g_server:broadcastEvent(self, false, connection, self.object)
    end
end

function ExtendAxleEvent.sendEvent(object, dir, noEventSend)
    if noEventSend == nil or noEventSend == false then
        if g_server ~= nil then
            g_server:broadcastEvent(ExtendAxleEvent:new(object, dir), nil, nil, object)
        else
            g_client:getServerConnection():sendEvent(ExtendAxleEvent:new(object, dir))
        end
    end
end
