--
-- OptiDrill
--
-- Authors: Wopster
-- Description: Enables you to make use of depth control and some hardcoded features for the kaweco opti drill
--
-- Copyright (c) Wopster, 2017

OptiDrill = {}
OptiDrill.STEP_SIZE = 0.25

function OptiDrill.prerequisitesPresent(specializations)
    if not SpecializationUtil.hasSpecialization(AnimatedVehicle, specializations) then
        print("Warning: Specialization OptiDrill needs the specialization AnimatedVehicle")
        return false
    end

    return true
end

function OptiDrill:preLoad(savegame)
    self.changeDepth = OptiDrill.changeDepth
end

function OptiDrill:load(savegame)
    self.animationNameManureflow = getXMLString(self.xmlFile, "vehicle.optiDrill#manureFlowAnimationName")

    local animationName = getXMLString(self.xmlFile, "vehicle.heightAdjustment#animationName")
    if animationName == nil then
        print("HeightAdjustment error: animationName cannot be empty!")
        return
    end

    self.heightAdjustment = {
        playerDistance = math.huge,
        playerMinDistance = 3,
        depthState = 1,
        onChange = false,
        onReset = false,
        animationName = animationName,
        inputBinding = InputBinding.IMPLEMENT_EXTRA2
    }

    local input = getXMLString(self.xmlFile, "vehicle.heightAdjustment#inputBinding")
    if input ~= nil and InputBinding[input] ~= nil then
        self.heightAdjustment.inputBinding = InputBinding[input]
    end
end

function OptiDrill:postLoad(savegame)
    if self.heightAdjustment.animationName ~= nil and self.playAnimation ~= nil then
        self:playAnimation(self.heightAdjustment.animationName, -1, nil, true)
    end

    if savegame ~= nil and not savegame.resetVehicles then
        local depthState = getXMLInt(savegame.xmlFile, savegame.key .. "#depthState")

        if depthState ~= nil then
            self:changeDepth(depthState, true)
        end
    end
end

function OptiDrill:delete()
end

function OptiDrill:getSaveAttributesAndNodes(nodeIdent)
    local attributes

    if self.heightAdjustment.depthState ~= nil then
        attributes = ('depthState="%s"'):format(self.heightAdjustment.depthState)
    end

    return attributes, nil
end

function OptiDrill:mouseEvent(...)
end

function OptiDrill:keyEvent(...)
end

function OptiDrill:update(dt)
    if g_currentMission.controlPlayer and g_currentMission.player ~= nil then
        if self:getToggledFoldDirection() > 0 then
            if self.attacherVehicle ~= nil then
                local implementIndex = self.attacherVehicle:getImplementIndexByObject(self)

                if implementIndex ~= nil then
                    local px, py, pz = getWorldTranslation(g_currentMission.player.rootNode)
                    local x, y, z = getWorldTranslation(self.rootNode)
                    local distance = Utils.vector3Length(px - x, py - y, pz - z)
                    self.heightAdjustment.playerDistance = math.huge

                    if distance < self.heightAdjustment.playerDistance and distance < self.heightAdjustment.playerMinDistance then
                        local jointDescIndex = self.attacherVehicle.attachedImplements[implementIndex].jointDescIndex
                        local joint = self.attacherVehicle.attacherJoints[jointDescIndex]

                        if not joint.moveDown then
                            self.heightAdjustment.onChange = true
                        else
                            self.heightAdjustment.onChange = false
                            g_currentMission:addExtraPrintText(g_i18n:getText("info_heightAdjustmentWarning"):format(self.typeDesc))
                        end

                        self.heightAdjustment.playerDistance = distance
                    else
                        self.heightAdjustment.onChange = false
                    end

                    if self.heightAdjustment.onChange then
                        if InputBinding.hasEvent(self.heightAdjustment.inputBinding) then
                            if not self.getIsAnimationPlaying(self, self.heightAdjustment.animationName) then
                                self:changeDepth(self.heightAdjustment.depthState + 1)
                            end
                        end

                        if g_i18n:hasText("action_heightAdjustment") then
							local depthState = self.heightAdjustment.depthState							
                            g_currentMission:addHelpButtonText(g_i18n:getText("action_heightAdjustment"):format(depthState < 4 and depthState + 1 or depthState - (depthState - 1)), self.heightAdjustment.inputBinding)
                        end
                    end
                end
            end
        end
    end
end

function OptiDrill:updateTick(dt)
    if self.attacherVehicle ~= nil then
        local implementIndex = self.attacherVehicle:getImplementIndexByObject(self)

        if implementIndex ~= nil then
            local jointDescIndex = self.attacherVehicle.attachedImplements[implementIndex].jointDescIndex
            local joint = self.attacherVehicle.attacherJoints[jointDescIndex]

            if joint ~= nil then
                local dir = (self.attacherVehicle.getIsTurnedOn ~= nil and self.attacherVehicle:getIsTurnedOn() and joint.moveDown) and 1 or -1

                self:playAnimation(self.animationNameManureflow, dir, nil, true)
            end
        end
    end
end

function OptiDrill:draw()
    if g_i18n:hasText("info_heightAdjustment") then
        g_currentMission:addExtraPrintText(g_i18n:getText("info_heightAdjustment"))
    end
end

function OptiDrill:onDetach(vehicle)
    if self:getAnimationTime(self.animationNameManureflow) > 0 then
        self:playAnimation(self.animationNameManureflow, -1, nil, true)
    end

    setVisibility(self.groundParticleAnimations[1].mappings[1].animNode, false)
end

---
-- @param depthState
-- @param noEventSend
--
function OptiDrill:changeDepth(depthState, noEventSend)
    ChangeDepthEvent.sendEvent(self, depthState, noEventSend)

    if self.heightAdjustment.animationName ~= nil then
        if depthState <= 4 then
            self:setAnimationStopTime(self.heightAdjustment.animationName, depthState / 4)
            self:playAnimation(self.heightAdjustment.animationName, 1, (depthState / 4) - OptiDrill.STEP_SIZE, true)
        else
            depthState = 1
            self:playAnimation(self.heightAdjustment.animationName, -1, 1, true)
        end
    end

    self.heightAdjustment.depthState = depthState
end

ChangeDepthEvent = {}
ChangeDepthEvent_mt = Class(ChangeDepthEvent, Event)

InitEventClass(ChangeDepthEvent, "ChangeDepthEvent")

function ChangeDepthEvent:emptyNew()
    local event = Event:new(ChangeDepthEvent_mt)
    return event
end

function ChangeDepthEvent:new(object, depthState)
    local event = ChangeDepthEvent:emptyNew()

    event.object = object
    event.depthState = depthState

    return event
end

function ChangeDepthEvent:readStream(streamId, connection)
    self.object = readNetworkNodeObject(streamId)
    self.depthState = streamReadInt8(streamId)
    self:run(connection)
end

function ChangeDepthEvent:writeStream(streamId, connection)
    writeNetworkNodeObject(streamId, self.object)
    streamWriteInt8(streamId, self.depthState)
end

function ChangeDepthEvent:run(connection)
    self.object:changeDepth(self.depthState, true)

    if not connection:getIsServer() then
        g_server:broadcastEvent(self, false, connection, self.object)
    end
end

function ChangeDepthEvent.sendEvent(object, depthState, noEventSend)
    if noEventSend == nil or noEventSend == false then
        if g_server ~= nil then
            g_server:broadcastEvent(ChangeDepthEvent:new(object, depthState), nil, nil, object)
        else
            g_client:getServerConnection():sendEvent(ChangeDepthEvent:new(object, depthState))
        end
    end
end
