--
-- TirePressure
--
-- Authors: Wopster
-- Description: Tire pressure specialization for vehicles
--
-- Copyright (c) Wopster, 2017

TirePressure = {}

TirePressure.INCREASE = 0.25
TirePressure.DEFAULT_MAX = 4
TirePressure.DEFAULT_MIN = 2
TirePressure.DEFAULT_SPEED = 0.0001
TirePressure.DEFAULT_EMPTY = 0.55
TirePressure.DEFAULT_FILL = 0.8

local PARAM_MORPH = "morphPosition"

function TirePressure.prerequisitesPresent(specializations)
    return true
end

function TirePressure:preLoad(savegame)
    self.setTirePressureState = TirePressure.setTirePressureState
    self.getTirePressureInflation = TirePressure.getTirePressureInflation
    self.setTirePressureInflation = TirePressure.setTirePressureInflation
    self.getIsValveOpen = TirePressure.getIsValveOpen
end

function TirePressure:load(savegame)
    self.tirePressure = {
        maxPressure = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.tirePressure#maxPressure"), TirePressure.DEFAULT_MAX),
        minPressure = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.tirePressure#minPressure"), TirePressure.DEFAULT_MIN),
        emptySpeedFactor = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.tirePressure#emptySpeedFactor"), TirePressure.DEFAULT_EMPTY),
        fillSpeedFactor = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.tirePressure#fillSpeedFactor"), TirePressure.DEFAULT_FILL),
        pressureSpeed = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.tirePressure#pressureSpeed"), TirePressure.DEFAULT_SPEED),
        isValveOpen = false,
        updatePressure = false,
        stopSoundActive = false,
    }

    self.tirePressure.pressure = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.TirePressure#default"), self.tirePressure.maxPressure)
    self.tirePressure.inflactionPressure = self.tirePressure.pressure
	self.inputBindingAirPressure = InputBinding.IMPLEMENT_EXTRA3
		
    local input = getXMLString(self.xmlFile, "vehicle.tirePressure#inputBinding")
    if input ~= nil and InputBinding[input] ~= nil then
        self.inputBindingAirPressure = InputBinding[input]
    end

    if self.isClient then
        local emptySampleNode = Utils.indexToObject(self.components, Utils.getNoNil(getXMLString(self.xmlFile, "vehicle.tirePressure.emptySound#linkNode"), "0>"))
        local fillSampleNode = Utils.indexToObject(self.components, Utils.getNoNil(getXMLString(self.xmlFile, "vehicle.tirePressure.fillSound#linkNode"), "0>"))

        self.sampleEmptyAirSound = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.tirePressure.emptySound", "$data/maps/sounds/siloFillSound.wav", self.baseDirectory, emptySampleNode)
        self.sampleFillAirSound = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.tirePressure.fillSound", "$data/maps/sounds/siloFillSound.wav", self.baseDirectory, fillSampleNode)
        self.sampleStopAirSound = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.tirePressure.stopSound", "$data/sounds/technicalAccessories/compressor_stop.wav", self.baseDirectory, fillSampleNode)
    end
end

function TirePressure:postLoad(savegame)
    if savegame ~= nil and not savegame.resetVehicles then
        local tirePressure = getXMLFloat(savegame.xmlFile, savegame.key .. "#tirePressure")

        if tirePressure ~= nil then
            self.tirePressure.pressure = tirePressure
            self:setTirePressureInflation()
        end
    end

    self:setTirePressureInflation()
end

function TirePressure:delete()
    if self.isClient then
        SoundUtil.deleteSample(self.sampleEmptyAirSound)
        SoundUtil.deleteSample(self.sampleFillAirSound)
        SoundUtil.deleteSample(self.sampleStopAirSound)
    end
end

function TirePressure:readStream(streamId, connection)
    self:setTirePressureState(streamReadBool(streamId), true)
end

function TirePressure:writeStream(streamId, connection)
    streamWriteBool(streamId, self.tirePressure.isValveOpen)
end

function TirePressure:getSaveAttributesAndNodes(nodeIdent)
    local attributes = ('tirePressure="%s"'):format(self.tirePressure.pressure)
    return attributes, nil
end

function TirePressure:mouseEvent(...)
end

function TirePressure:keyEvent(...)
end

function TirePressure:update(dt)
    if self:getIsActive() then
        if self:getIsActiveForInput() and not self:hasInputConflictWithSelection() then
            if InputBinding.hasEvent(self.inputBindingAirPressure) then
                self:setTirePressureState(not self.tirePressure.isValveOpen)
            end
        end
    end
end

function TirePressure:updateTick(dt)
    if self.tirePressure.updatePressure then
        local pressureSpeed = self.tirePressure.pressureSpeed * dt
        local pressureSpeedFactor = self:getIsValveOpen() and self.tirePressure.emptySpeedFactor or self.tirePressure.fillSpeedFactor

        if self:getIsValveOpen() then
            self.tirePressure.pressure = Utils.clamp(self.tirePressure.pressure - (pressureSpeed * pressureSpeedFactor), self.tirePressure.minPressure, self.tirePressure.maxPressure)

            if self.isClient then
                if self.tirePressure.minPressure < self.tirePressure.pressure then
                    SoundUtil.play3DSample(self.sampleEmptyAirSound)
                else
                    self.tirePressure.stopSoundActive = true
                    self.tirePressure.updatePressure = false

                    SoundUtil.stop3DSample(self.sampleEmptyAirSound)
                end

                if TirePressure.is3DSamplePlaying(self.sampleFillAirSound) then
                    SoundUtil.stop3DSample(self.sampleFillAirSound)
                end
            end
        else
            self.tirePressure.pressure = Utils.clamp(self.tirePressure.pressure + (pressureSpeed * pressureSpeedFactor), self.tirePressure.minPressure, self.tirePressure.maxPressure)

            if self.isClient then
                if self.tirePressure.maxPressure > self.tirePressure.pressure then
                    SoundUtil.play3DSample(self.sampleFillAirSound)
                else
                    self.tirePressure.stopSoundActive = true
                    self.tirePressure.updatePressure = false

                    SoundUtil.stop3DSample(self.sampleFillAirSound)
                end

                if TirePressure.is3DSamplePlaying(self.sampleEmptyAirSound) then
                    SoundUtil.stop3DSample(self.sampleEmptyAirSound)
                end
            end
        end

        self:setTirePressureInflation()
    end

    if self.isClient then
        if not self.tirePressure.updatePressure and self.tirePressure.stopSoundActive and not TirePressure.is3DSamplePlaying(self.sampleStopAirSound) then
            SoundUtil.play3DSample(self.sampleStopAirSound)
        elseif self.tirePressure.stopSoundActive and TirePressure.is3DSamplePlaying(self.sampleStopAirSound) then
            self.tirePressure.stopSoundActive = false
            SoundUtil.stop3DSample(self.sampleStopAirSound)
        end
    end
end

function TirePressure:draw()
    if self.tirePressure.updatePressure then
        if g_i18n:hasText("info_tireFilling") and g_i18n:hasText("info_tireEmptying") and g_i18n:hasText("info_tireCurrentPressure") then
            local targetText = not self:getIsValveOpen() and g_i18n:getText("info_tireFilling") or g_i18n:getText("info_tireEmptying")

            g_currentMission:addExtraPrintText(g_i18n:getText("info_tireCurrentPressure"):format(targetText, self.tirePressure.pressure))
        end
    else
        if g_i18n:hasText("action_tireChangeTarget") then
            local targetTirePressure = not self:getIsValveOpen() and self.tirePressure.minPressure or self.tirePressure.maxPressure

            g_currentMission:addHelpButtonText(g_i18n:getText("action_tireChangeTarget"):format(targetTirePressure), self.inputBindingAirPressure)
        end
    end
end

---
-- @param state
-- @param noEventSend
--
function TirePressure:setTirePressureState(state, noEventSend)
    SetTirePressureStateEvent.sendEvent(self, state, noEventSend)

    self.tirePressure.isValveOpen = state
    self.tirePressure.updatePressure = true
end

---
--
function TirePressure:getIsValveOpen()
    return self.tirePressure.isValveOpen
end

---
--
function TirePressure:getTirePressureInflation()
    return self.tirePressure.inflactionPressure
end

---
--
function TirePressure:setTirePressureInflation()
    self.tirePressure.inflactionPressure = math.abs(1 - Utils.clamp(self.tirePressure.pressure / self.tirePressure.maxPressure, 0, 1))
end

---
-- Giants didn't create this function..
-- @param sample
-- @param offset
--
function TirePressure.is3DSamplePlaying(sample, offset)
    if sample ~= nil and sample.sound3D ~= nil then
        offset = Utils.getNoNil(offset, 0)
        local source = getAudioSourceSample(sample.sound3D)

        if isSamplePlaying(source) and getSamplePlayOffset(source) < getSampleDuration(source) - offset then
            return true
        end
    end

    return false
end

---
-- @param self
-- @param wheel
-- @param x
-- @param y
-- @param z
-- @param xDrive
-- @param suspensionLength
--
function TirePressure.updatePressureWheelGraphics(self, wheel, x, y, z, xDrive, suspensionLength)
    if self.getTirePressureInflation ~= nil and self:getTirePressureInflation() ~= nil then
        if wheel.wheelTire ~= nil then
            local tireTypeCrawler = WheelsUtil.getTireType("crawler")

            if wheel.tireType ~= tireTypeCrawler then
                local x, y, z, _ = getShaderParameter(wheel.wheelTire, PARAM_MORPH)
                local deformation = Utils.clamp((wheel.deltaY + 0.04 - wheel.netInfo.suspensionLength) * (TirePressure.INCREASE + self:getTirePressureInflation()), 0, wheel.maxDeformation)

                -- Redo the shader morph for better graphical display.. could have just clamped the maxDeformation value but that doesn't really give the correct visual feeling.
                setShaderParameter(wheel.wheelTire, PARAM_MORPH, x, y, z, deformation, false)

                if wheel.additionalWheels ~= nil then
                    for _, additionalWheel in pairs(wheel.additionalWheels) do
                        local x, y, z, _ = getShaderParameter(additionalWheel.wheelTire, PARAM_MORPH)
                        setShaderParameter(additionalWheel.wheelTire, PARAM_MORPH, x, y, z, deformation, false)
                    end
                end

                suspensionLength = wheel.netInfo.suspensionLength + deformation
            end
        end
    end
end

WheelsUtil.updateWheelGraphics = Utils.appendedFunction(WheelsUtil.updateWheelGraphics, TirePressure.updatePressureWheelGraphics)

SetTirePressureStateEvent = {}
SetTirePressureStateEvent_mt = Class(SetTirePressureStateEvent, Event)

InitEventClass(SetTirePressureStateEvent, "SetTirePressureStateEvent")

function SetTirePressureStateEvent:emptyNew()
    local event = Event:new(SetTirePressureStateEvent_mt)
    return event
end

function SetTirePressureStateEvent:new(object, state)
    local event = SetTirePressureStateEvent:emptyNew()

    event.object = object
    event.state = state

    return event
end

function SetTirePressureStateEvent:readStream(streamId, connection)
    self.object = readNetworkNodeObject(streamId)
    self.state = streamReadBool(streamId)

    self:run(connection)
end

function SetTirePressureStateEvent:writeStream(streamId, connection)
    writeNetworkNodeObject(streamId, self.object)
    streamWriteBool(streamId, self.state)
end

function SetTirePressureStateEvent:run(connection)
    self.object:setTirePressureState(self.state, true)

    if not connection:getIsServer() then
        g_server:broadcastEvent(self, false, connection, self.object)
    end
end

function SetTirePressureStateEvent.sendEvent(object, state, noEventSend)
    if noEventSend == nil or noEventSend == false then
        if g_server ~= nil then
            g_server:broadcastEvent(SetTirePressureStateEvent:new(object, state), nil, nil, object)
        else
            g_client:getServerConnection():sendEvent(SetTirePressureStateEvent:new(object, state))
        end
    end
end
